# Environment variable names
ENV_FURIOSA_LOG_DIR = "FURIOSA_LOG_DIR"
ENV_FURIOSA_LOG_LEVEL = "FURIOSA_LOG_LEVEL"
ENV_NPU_DEVNAME = "NPU_DEVNAME"
ENV_XDG_STATE_HOME = "XDG_STATE_HOME"

# Default NPU device
DEFAULT_DEVNAME = "npu0pe0"
